#include "master.h"

typedef struct {
     KeySym *row_syms[5];
     int row_counts[5];
     int rows;
     int keysyms_per_row;
     char *filename;
} KEYBOARD;

#define NO_KEYBOARD                 ((KEYBOARD *) NULL)
#define NULLIFY(variable, type)     variable = (type *)NULL
#define KEYBOARD_KEYSYMS_PER_KEY    3    /* Norm, caps, AltGr */

void destroy_keyboard(KEYBOARD *keyb)
{
     register int x=0;
     for(x=0; x<5; x++)
     {
	  if(keyb->row_syms[x])
	  {
	       g_free_(keyb->row_syms[x]);
	       NULLIFY(keyb->row_syms[x], KeySym);
	  } /* End if */
     } /* End for */

     if(keyb->filename)
     {
#ifdef DEBUGGING
	  fprintf(Q,"Freeing KEYBOARD filename.\n"); fflush(Q);
#endif /* DEBUGGING */
	  g_free_(keyb->filename);
	  NULLIFY(keyb->SYMS, char);
     } /* End if */

     if(keyb)
     {
#ifdef DEBUGGING
	  fprintf(Q,"Freeing KEYBOARD structure.\n"); fflush(Q);
#endif
	  g_free_(keyb);
     }

#ifdef DEBUGGING
     fprintf(Q,"NULLIFYING KEYBOARD structure.\n"); fflush(Q);
#endif
     NULLIFY(keyb, KEYBOARD);
} /* End destroy_keyboard() */

KEYBOARD *readdef(char *filename)
{
     KEYBOARD *keyb = NO_KEYBOARD;
     FILE *fp = fopen(filename, "r");
     register int x=0, y=0, z=0;
     int line_size = 1024;
     int line_number = 0;
     char linedata[line_size];
     char **tokens;
     char *syms[3];
     long keysyms_needed = 0;

     if(!filename)
     {
	  fprintf(stderr,"Readdef error:  NULL filename.\n");
	  fflush(stderr);
	  return(NO_KEYBOARD);
     } /* End if */

     fp = fopen(filename, "r");

     if(!fp)
     {
	  fprintf(stderr,"Error:  Couldn't open %s for reading: %s\n",
		  filename, g_strerror(errno)); 
	  fflush(stderr);
	  return(NO_KEYBOARD);
     }

     for(x=0; x<5; x++)  /* Read 5 lines worth of data */
     {
	  /* Read the number of tokens in this chunk. */
	  linedata[0] = '\0';
	  FILE_readline(fp, &linedata[0], line_size);
	  line_number++;
	  
	  if(!linedata[0])
	  {
	       fprintf(stderr,"Error:  Premature EOF file %s row %d.\n",
		       filename, x); fflush(stderr);
	       return(NO_KEYBOARD);
	  }

	  if((keyb->row_counts[x] = atoi(linedata)) <= 0)
	  {
	       fprintf(stderr,"Illegal token count of %d row %d.\n",
		       keyb->row_counts[x], x); fflush(stderr);
	       return(NO_KEYBOARD);
	  } /* End if */

	  for(y=0; y<keyb->row_counts[x]; y++)
	  {
#ifdef DEBUGGING
	       fprintf(Q,"Reading chunk %d of keyboard row %d (%d max).\n",
		       y, x, keyb->row_counts[x]); fflush(Q);
#endif /* DEBUGGING */

	       linedata[0] = '\0';
	       FILE_readline(fp, &linedata[0], line_size);
	       line_number++;
	       
	       if(!linedata[0])
	       {
		    fprintf(stderr,"Error:  Premature EOF file %s row %d.\n",
			    filename, x); fflush(stderr);
		    return(NO_KEYBOARD);
	       } /* End if */
	       
	       tokens = g_strsplit(linedata, " \t\n\r", -1);
	  
	       syms[0] = syms[1] = syms[2] = (KeySym *)NULL;
	       
	       for(z=0; z<3; z++)
	       {
		    syms[z] = *tokens++;
		    if(!syms[z])
		    {
			 fprintf(Q,"Error file %s, line %d:  Tokens missing\n",
				 filename, line_number); fflush(Q);
			 return(NO_KEYBOARD);
		    } /* End if */
	       } /* End for */

	       /* Now insert the keysyms into the structure */
	       keyb->row_syms[x] = g_new0(KeySym, 4);

	       for(z=0; z<3; z++)
	       {
#ifdef DEBUGGING
		    fprintf(Q,"Copying %s into row_syms[%d][%d].\n",
			    syms[z], x, z); fflush(Q);
#endif
		    keyb->row_syms[x][z] = XStringToKeysym(syms[z]);
	       } /* End for */

#ifdef DEBUGGING
	       fprintf(Q,"Freeing multitokens.\n"); fflush(Q);
#endif
	       g_strfreev(tokens);
	  } /* End for */
     } /* End for */

     fclose(fp);
     
     keyb->filename = g_strdup_(filename);
     keyb->rows     = 5;
     return(keyb);
} /* End readdef() */







