/* CALLBACKS.C 
 * This file is for use with GTKeyboard written by David Allen
 * s2mdalle@titan.vcu.edu
 * http://opop.nols.com/
 * 
 * This file is released under the terms of the GNU GPL.  Please see 
 * http://www.gnu.org for more details.
 */

#define CALLBACKS_C 1
#define APP_C 1   /* So this file can get at app.c's data in master.h */
#include "master.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <errno.h>
#include <signal.h>
#include <ctype.h>

#define MODE_WRITE
#define DEBUGGING

/* Do not allow both MODE_WRITE and MODE_APPEND to be defined.  If both are
 * defined, default to MODE_WRITE.  The End Of Earth As We Know It (TM) occurs
 * (or maybe just a compile error) MODE_WRITE is spiffier.
 */
#ifdef MODE_APPEND
#  ifdef MODE_WRITE
#    undef MODE_APPEND
#  endif /* MODE_WRITE */
#endif /* MODE_APPEND */

#ifndef __USE_GNU
#  define __USE_GNU      
#endif /* __USE_GNU */
#define MESSAGE_WIDTH  200
#define MESSAGE_HEIGHT 100

/* Static function prototypes */
static void gtkeyboard_mem_cleanup(void);

/* Copy, paste, cut functions */
void CutText(GtkWidget *w, gpointer data)
{
     gtk_editable_cut_clipboard(GTK_EDITABLE(GUI.main_output_text));
     chocolate("Text cut to the clipboard.\n");
     return;
}

void CopyText(GtkWidget *w, gpointer data)
{
     gtk_editable_copy_clipboard(GTK_EDITABLE(GUI.main_output_text));
     chocolate("Text copied to the clipboard.\n");
     return;
}

void PasteText(GtkWidget *w, gpointer data)
{
     gtk_editable_paste_clipboard(GTK_EDITABLE(GUI.main_output_text));
     chocolate("Text pasted from the clipboard.\n");
     return;
}

long sizeof_text_widget(GtkWidget *widget)
{
     char *buffer = gtk_editable_get_chars(GTK_EDITABLE(widget), 0, -1);
     long rep = (long)strlen(buffer) * sizeof(char);
     g_free_(buffer);
     return(rep);
} /* End sizeof_text_widget */

/* Status report just prints out what options.CAPS_LOCK and options.SHIFT are 
 * to the status widget among other things.
 */
void status_report(GtkWidget *widget, gpointer data)
{
     /* Basically just strcat a bunch of crap together and throw it up
      * to the screen.
      */
     
     char temp[2048];
     char on[]="ON ", 
	  off[]="OFF ", 
	  widgetwarning[] = "You do not have the main text widget enabled.\nThat's fine, but it also means that you can't use GTKeyboard as a text editor.\nI humbly suggest that you select another output window to type into instead of just hitting keys, which will most assuredly do nothing with this configuration.\n";
     sprintf(temp,"----====< STATUS >====----");
     if(options.ASK_SAVE_ON_EXIT)
	  strcat(temp,"\nYou will be asked if you want to save on exit.");
     else
	  strcat(temp,"\nYou will not be asked if you want to save on exit.");

     strcat(temp,"\nSTATUS LOGGING is ");
     if(options.STATUS_LOGGING==1)
	  strcat(temp,on);
     else
	  strcat(temp,off);
     
     strcat(temp,"\nWORD WRAP is ");
     if(options.WORD_WRAP)
	  strcat(temp,on);
     else
	  strcat(temp,off);

     if(options.INFO_POPUPS)
	  strcat(temp,"\nYou will be shown confirmation popup windows.");

     strcat(temp,"\nNUM LOCK is ");
     if(options.NUMLOCK)
	  strcat(temp,on);
     else
	  strcat(temp,off);

     strcat(temp,"\nOur workingfile is ");
     if(!options.workingfile)
	  strcat(temp,"(None)");
     else
	  strcat(temp,options.workingfile);

     strcat(temp,"\n");
     strcat(temp,"Our text space has ");
     sprintf(temp,"%s %ld ",temp,sizeof_text_widget(GUI.main_output_text));
     strcat(temp,"characters in it\n");

     gtkeyboard_message(1,temp);

     if(!options.SHOW_ALL_WINDOWS)
	  gtkeyboard_message(1,widgetwarning);

     if(options.INFO_POPUPS)
	  annoying_popup(temp);

     return;
} /* End status report */

void smack(GtkWidget *w, gpointer data)
{
     GtkWidget *prisoner = (GtkWidget *)data;
     /* Get it out of our faces */
     gtk_widget_destroy(w);
     gtk_widget_destroy(prisoner);

     prisoner = (GtkWidget *)NULL;
     w        = (GtkWidget *)NULL;
     return;
} /* End smack */

/* Pop up a save file as dialog */
void save_file_as(GtkWidget *emitter, gpointer data)
{
     GtkWidget *popup  = gtk_dialog_new();
     GtkWidget *label  = gtk_label_new("Save file as:");
     GtkWidget *fname  = gtk_entry_new_with_max_length(50);
     GtkWidget *ok     = create_ok_button(GTK_SIGNAL_FUNC(
	  gtkeyboard_grab_filename_save), fname);
     GtkWidget *cancel = create_cancel_button(popup);
     GtkWidget *hsep   = gtk_hseparator_new();

     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(popup)->vbox), label,
			FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(popup)->vbox), hsep,
			FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(popup)->vbox), fname,
			FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(popup)->action_area), ok,
			FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(popup)->action_area), cancel,
			FALSE, FALSE, 0);

     gtk_widget_show_all(popup);

} /* End save_file_as */

/* This is an "are you sure?" dialog for remapping the keyboard since
 * people might not want to actually do it. The ok button maps onto
 * gtkeyboard_remap_keyboard()
 */
void remap_keyboard_really(GtkWidget *emitter, gpointer data)
{
     GtkWidget *popup = gtk_dialog_new();
     GtkWidget *ok;
     GtkWidget *cancel;
     GtkWidget *label;
     GtkWidget *label2;
     char txt[1024 * 5];
     char warn[512];

     ok = create_ok_button(GTK_SIGNAL_FUNC(gtkeyboard_remap_keyboard), NULL);

     sprintf(txt,"Would you like to remap the keyboard to the\n");
     
     if(options.keyboard->name)
     {
	  strcat(txt,"keyboard definition in the file:\n");
	  strcat(txt,options.keyboard->name);
	  strcat(txt,"?\n");
     }
     else
	  strcat(txt,"currently shown layout?\n");

     sprintf(warn,"THIS WILL CHANGE YOUR PHYSICAL KEYBOARD'S MAPPING.\n");

     /* Cancel button is a regular gtk_widget_destroy(popup) */
     cancel = create_cancel_button(popup);

     label = gtk_label_new(txt);
     label2 = gtk_label_new(warn);

     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(popup)->vbox), label,
			FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(popup)->vbox), label2,
			FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(popup)->action_area), ok,
			FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(popup)->action_area), cancel,
			FALSE, FALSE, 0);

     gtk_window_set_position(GTK_WINDOW(popup), GTK_WIN_POS_CENTER);

     gtk_widget_show_all(popup);
} /* End remap_keyboard_really() */

/* Grab the changed filename from the save_as dialog and save the file
 * that the user specified.
 */
void gtkeyboard_grab_filename_save(GtkWidget *emitter, gpointer data)
{
     GtkWidget *entrywidget = (GtkWidget *)data;
     char *new_filename = (char *)NULL;

     if(!GTK_IS_ENTRY(entrywidget))
     {
	  gtkeyboard_error(3,"Internal GTKeyboard error.  Please report ",
			   "this error along with what you were doing at ",
			   "the time.\n");
	  return;
     } /* End if */

     new_filename = gtk_entry_get_text(GTK_ENTRY(entrywidget));

     gtk_widget_destroy(entrywidget->parent->parent);

     if(options.workingfile)
     {
	  g_free_(options.workingfile);
	  options.workingfile = NULL;
     } /* End if */

     options.workingfile = g_strdup_(new_filename);
     save_output_text(NULL, NULL);

     gtkeyboard_message(1,"Changed filename and saved.\n");
     return;
} /* End gtkeyboard_grab_filename_save */

void save_output_text(GtkWidget *w, gpointer data)
{
     char *buffer = (char *)NULL;
     FILE *fp;
     int nchars;
     GtkWidget *foobar = (GtkWidget *)data;

     if(!options.workingfile || options.workingfile == (char *)NULL)
     {
	  gtkeyboard_error(1,"No filename specified.\n");
	  save_file_as(NULL, NULL);
	  return;
     } /* End if */

     if(!foobar)
	  foobar = GUI.main_output_text;

     /* Mode append or write can be specified with symbolics at the top
      * of this file.  I highly recommend write, as most people will think
      * it's pretty weird to do appending.  Still, somebody asked for it..
      */

#ifdef MODE_APPEND
     if((fp=fopen(options.workingfile, "a"))!=NULL)
     {
#endif /* MODE_APPEND */
#ifdef MODE_WRITE
     if((fp=fopen(options.workingfile, "w"))!=NULL)
     {
#endif /* MODE_WRITE */

	  /* Freeze so user can't edit while we're saving */
	  gtk_text_freeze(GTK_TEXT(foobar));
	  
	  buffer = gtk_editable_get_chars (GTK_EDITABLE (foobar),
					   (gint) 0,
					   (gint) gtk_text_get_length(
						GTK_TEXT (foobar)));
	  
	  nchars = fwrite (buffer, sizeof (char), strlen (buffer), fp);

	  if(nchars!=strlen(buffer))
	       save_output_text_error("Couldn't write enough bytes!");
	  
	  g_free_(buffer);
	  buffer = NULL;
	  
	  /* Unfreeze so editing can keep going */
	  gtk_text_thaw(GTK_TEXT(foobar));
	  fflush(fp);
	  fclose(fp);
	  save_output_text_successful(options.workingfile);
     } /* End if */
     else
     {
	  /* Put up a popup with the errno error in it. */
	  save_output_text_error(g_strerror(errno));
     } /* End else */
     
     if(buffer)
     {
	  g_free_(buffer);
     } /* End if */

     return;
} /* End save_output_text */

void save_output_text_error(char *error)
{
     char buffer[1024];

     sprintf(buffer,"Error saving %s:  %s",options.workingfile,error);

     if(options.INFO_POPUPS)
	  annoying_popup(buffer);

     chocolate(buffer);
     chocolate("\n\n");
     return;
} /* End save_output_text_error */

/* Save_output_text_successful - happens when we saved a file */
void save_output_text_successful(char *filename)
{
     char buffer[1024];   /* Kudos to you if you have a filename to overflow
			   * this.  :)
			   */
     sprintf(buffer,"File save successful:  %s",filename);

     if(options.INFO_POPUPS)
	  annoying_popup(buffer);

     chocolate(buffer);
     chocolate("\n\n");
     return;
} /* End save_output_text_success */

/* Showmenu */
void Showmenu(GtkWidget *w, gint input)
{
     char message[100];
     
     switch(input)
     {
     case FILEM:
	  strcpy(message,"FILE menu torn off.\n");
	  break;
     case EDITM:
	  strcpy(message,"EDIT menu torn off.\n");
	  break;
     case OPTIONSM:
	  strcpy(message,"OPTIONS menu torn off.\n");
	  break;
     case MISCM:
	  strcpy(message,"MISC menu torn off.\n");
	  break;
     case HELPM:
	  strcpy(message,"HELP menu torn off.\n");
	  break;
     case HTMLM:
	  strcpy(message,"HTML menu torn off.\n");
	  break;
     case SHORTCUTSM:
	  strcpy(message,"SHORTCUTS menu torn off.\n");
	  break;
     default:
	  strcpy(message,"MENU torn off.\n");
	  break;
     } /* End switch */

     chocolate(message);
     return;
} /* End showmenu */

void toggle_double_data(void)
{
     char on[] = 
	  "All keys will be sent to the textarea and the redirect window.\n";
     char off[] = 
	  "Keys will only be sent to the window you are working with.\n";

     if(ISON(options.SEND_TO_BOTH_WINDOWS))
     {
	  chocolate(off);
	  options.SEND_TO_BOTH_WINDOWS = OFF;
	  if(options.INFO_POPUPS)
	       annoying_popup(off);
     } /* End if */
     else
     {
	  chocolate(on);
	  options.SEND_TO_BOTH_WINDOWS = ON;
	  if(options.INFO_POPUPS)
	       annoying_popup(on);
     } /* End else */
} /* End toggle_double_data() */

/* Self explanatory */
void toggle_ask_save_on_exit(void)
{
     char on[]  = "You will be asked if you want to save on exit.\n";
     char off[] = "You will not be asked if you want to save on exit.\n";

     if(ISON(options.ASK_SAVE_ON_EXIT))
     {
	  chocolate(off);
	  options.ASK_SAVE_ON_EXIT=OFF;
	  if(options.INFO_POPUPS)
	       annoying_popup(off);
     } /* End if */
     else
     {
	  chocolate(on);
	  options.ASK_SAVE_ON_EXIT=ON;
	  if(options.INFO_POPUPS)
	       annoying_popup(on);
     } /* End else */
     
     return;
} /* End toggle_ask_save_on_exit */

/* Turn status logging on or off - only saves at exit */
void status_logging_toggle(void)
{
     char on[]  = "Status logging is ON\n";
     char off[] = "Status logging is OFF\n";

     if(ISON(options.STATUS_LOGGING))
     {
	  options.STATUS_LOGGING=OFF;
	  chocolate(off);
	  
	  if(options.INFO_POPUPS)
	       annoying_popup(off);

     } /* End if */
     else 
     {
	  options.STATUS_LOGGING=ON;
	  chocolate(on);

	  if(options.INFO_POPUPS)
	       annoying_popup(on);
     } /* End else */
     return;
} /* End status_logging_toggle */

gint CloseApp (GtkWidget *widget, gpointer data)
{
     /* Print a carriage return just to pretty things up so to speak */
     char filename[] = "gtkeyboard.status.txt";
     char *datafoo;
     FILE *fp;
     int flag = 0;
     
     if(GTK_IS_WIDGET((GtkWidget *)data))
	  gtk_widget_destroy((GtkWidget *)data);

     if(options.STATUS_LOGGING)
     {
	  fp = fopen(filename,"w");
	  if(fp==NULL)
	  {
	       /* Don't write info to status since we're about to die anyway.
		* wasteful.  :)
		*/
	       fprintf(Q,"Couldn't open %s for writing : %s\n",filename,
		       g_strerror(errno));
	       fprintf(Q,"Status log NOT written.\n");
	  }
	  else
	  {
	       datafoo = gtk_editable_get_chars(GTK_EDITABLE(GUI.status), 
						(gint) 0, (gint) -1);
	       if(!datafoo)
	       {
		    fprintf(Q,"GTKeyboard Error: CloseApp() data is NULL\n");
		    fflush(Q);
	       } /* End if */

	       /* Print to the file */
	       fprintf(fp,"%s",datafoo);
	       /* Flush the buffer */
	       fflush(fp);
	       /* Close it */
	       fclose(fp);
	       fprintf(stdout,
		       "Status log successfully written to %s\n",filename);
	       fflush(stdout);
	       /* Free that mem, yeehaw! :) */
	       g_free_(datafoo);
	  } /* End else */
     } /* End if options.STATUS_LOGGING */
     else
     {
	  fprintf(stdout,"Status logging wasn't enabled.  Log not written.\n");
	  fflush(stdout);
     }

     /* Do we need to save anything before we leave? */

     /* Watch out for that short circuit evaluation.  options.workingfile
      * COULD be NULL here, in which case strstr would cause us to 
      * segfault.
      */
     if(options.ASK_SAVE_ON_EXIT && options.workingfile &&
	!strstr(options.workingfile,"/dev/"))
     {
	  /* If we need to save, the confirm saving/give an option to save
	   * to a different file if needed.
	   */
	  fprintf(Q,"Ask save on exit was enabled.\n");
	  fflush(Q);

	  if(check_if_need_to_save(GUI.main_output_text, options.workingfile))
	       confirm_save_on_exit(options.workingfile);
	  else  
	  {
	       byebye((GtkWidget *)NULL, 0);
	       flag=1;
	  } /* End else */
     } /* End if */
     else
     {
#ifdef DEBUGGING
	  fprintf(Q,"Save on exit wasn't enabled ");
	  fprintf(Q,"or working with a device file.\n");
	  fflush(Q);
#endif /* DEBUGGING */
	  byebye((GtkWidget *)NULL, 0);
	  flag=1;
     } /* End else */

     if(flag)
     {
	  gtk_main_quit();
	  
	  MEM_REPORT;
	  gtkeyboard_mem_cleanup();
	  MEM_REPORT;

#ifndef PROD
	  g_mem_profile();
#endif /* !PROD */

	  MEM_REPORT;
	  flush_everything();
	  gtk_exit(0);
     } /* End if */

     /* this might not happen depending on flag */
     return(TRUE);
} /* End closeapp */

void gtkeyboard_signal_handler(void)
{
     fprintf(stderr,"Signal caught by GTKeyboard...shutting down.\n");
     fflush(stderr);

     /* Exit the program immediately. */
     byebye((GtkWidget *)NULL, 1);
     exit(1);     /* This should never be reached */
} /* End gtkeyboard_signal_handler */

/* Function exits the program and dies */
int byebye(GtkWidget *emitter, gint foo)
{
     /* Free up the mem from all the child widgets, and destroy everything
      * we can get our greasy fingers on
      */

     gtk_widget_unrealize(window);

     /* Frees all memory associated with all widget below */
     gtk_widget_destroy(window);

     if(foo==1)
     {
	  fprintf(Q,"Exiting GTKeyboard from signal emission/dialog box.\n");
	  fflush(Q);
	  gtk_main_quit(); 
     }
     
     flush_everything();
     
     if(foo==1)         /* Were we called without closeapp? */
     {
	  gtkeyboard_mem_cleanup();

#ifndef PROD
	  g_mem_profile();
#endif /* PROD */

	  MEM_REPORT;
	  flush_everything();
	  gtk_exit(0);
     } /* End if */
     
     /* May or may not happen */
     return(TRUE);
} /* End byebye */

/* Cleans up a bunch of memory that was allocated at different stages.
 * Make sure you don't call this unless you mean to exit really soon since
 * it destroys all relevant data structures
 *
 * If options.VERBOSE is ON, then it will print out what it is cleaning up
 * at the moment.  It also flushes, since earlier I messed up some of the
 * memory earlier, and it would segfault.  Flush before you clear the memory
 * so you'll know what caused the segfault.  Also, set everything to NULL
 * after you g_free() it because almost everything in GTKeyboard checks to 
 * see if the pointer is pointing somewhere valid before it tries to free
 * it and depending on your machine/compiler, it might still point somewhere
 * !=NULL after freed.
 */
static void gtkeyboard_mem_cleanup(void)
{
     int x = 0;

#ifndef PROD  /* If we're debugging assure these messages are seen */
     options.VERBOSE = ON;
#else /* defined PROD */
     options.VERBOSE = OFF;
#endif /* PROD */

     for(x=0; x<SHORTCUTS; x++)
     {
	  if(options.VERBOSE)
	  {
	       fprintf(Q,"Shortcut[%d]...",x);  
	       fflush(Q);
	  } /* End if */
	  if(options.shortcuts[x])
	  {
	       g_free_(options.shortcuts[x]);
	       options.shortcuts[x] = NULL;
	  } /* End if */
     } /* End for */
     
     for(x=0; x<MAX_CUSTOM_COLORS; x++)
     {
	  if(options.VERBOSE)
	  {
	       fprintf(Q,"Custom[%d]...",x); 
	       fflush(Q);
	  } /* End if */
	  if(GUI.custom_colors[x])
	  {
	       g_free_(GUI.custom_colors[x]);
	       options.shortcuts[x] = NULL;
	  } /* End if */
     } /* End for */

     MEMMSG("Destroying GUI.popup_menu.");    COND_DESTROY(GUI.popup_menu);

     MEMMSG("Cursor.");
     if(GUI.cursor)
     {
	  gdk_cursor_destroy(GUI.cursor);
	  GUI.cursor = NULL;
     } /* End if */

     /* Message about cleaning up then conditionally free a bunch of 
      * different items, yo
      */
     MEMMSG("szNeedle.");              CONDFREE(options.szNeedle);
     MEMMSG("extrafiles.");            CONDFREE(options.extrafiles);
     MEMMSG("home.");                  CONDFREE(options.home);
     MEMMSG("Redirect_window_name.");  CONDFREE(options.redirect_window_name);
     MEMMSG("Workingfile.");           CONDFREE(options.workingfile);
     MEMMSG("Extrafiles.");            CONDFREE(options.extrafiles);
     MEMMSG("Tempslot.");              CONDFREE(options.tempslot);
     MEMMSG("Fontname.");              CONDFREE(GUI.fontname);
     MEMMSG("KFontname.");             CONDFREE(GUI.kfontname);
     MEMMSG("Kfont.");                 CONDFREE(GUI.kfont);
     MEMMSG("Style.");                 CONDFREE(GUI.style);
     MEMMSG("KStyle.");                CONDFREE(GUI.kstyle);
     MEMMSG("Deflt.");                 CONDFREE(GUI.deflt);
     MEMMSG("Font.");                  CONDFREE(GUI.font);

     if(options.keyboard)
     {
	  MEMMSG("Options.Keyboard.");
	  options.keyboard = gtkeyboard_destroy_keyboard(options.keyboard);
	  options.keyboard = NO_KEYBOARD;
     }
     if(options.last_keyboard)
     {
	  MEMMSG("Options.Last_Keyboard.");
	  options.last_keyboard = 
	       gtkeyboard_destroy_keyboard(options.last_keyboard);
	  options.last_keyboard = NO_KEYBOARD;
     }

     MEMMSG("Done Freeing\n");
     flush_everything();

#ifndef PROD
     fprintf(Q,"gnew = %d\ngnew0 = %d\ngstrdup = %d\ngmalloc = %d\ngfree = %d",
	     options.gnew, options.gnew0, options.gstrdup, options.gmalloc,
	     options.gfree);
     fprintf(Q,"\n");
     fflush(Q);
#endif /* !PROD */
} /* End gtkeyboard_mem_cleanup */

void alt_gr_toggle(GtkWidget *w, gpointer data)
{
     if(ISOFF(options.ALT_GR))
     {
	  options.ALT_GR = ON;
	  chocolate(ALT_GR_IS_ON);
     }
     else 
     {
	  options.ALT_GR = OFF;
	  chocolate(ALT_GR_IS_OFF);
     }
} /* End alt_gr_toggle */

void control_toggle(GtkWidget *w, gpointer data)
{
     if(ISOFF(options.CONTROL))
     {
	  options.CONTROL = ON;
	  chocolate(CONTROL_IS_ON);
     } /* End if */
     else
     {
	  options.CONTROL = OFF;
	  chocolate(CONTROL_IS_OFF);
     } /* End else */
} /* End control_toggle */

void alt_toggle(GtkWidget *w, gpointer data)
{
     if(ISOFF(options.ALT))
     {
	  options.ALT = ON;
	  chocolate(ALT_IS_ON);
     } /* End if */
     else
     {
	  options.ALT = OFF;
	  chocolate(ALT_IS_OFF);
     } /* End else */
} /* End alt_toggle */

void capslock_toggle(GtkWidget *w, gpointer data)
{
     GtkWidget *foo = (GtkWidget *)data;

     if(options.redirect_window_name || options.SEND_TO_BOTH_WINDOWS)
	  send_redirect_a_keysym(XK_Caps_Lock);

     /* Whatever it currently is, swtich it */
     if(ISOFF(options.CAPS_LOCK))
     {	  
	  options.CAPS_LOCK=ON;
	  gtk_text_insert(GTK_TEXT(foo), NULL, NULL, NULL,
			  CAPS_LOCK_IS_ON, strlen(CAPS_LOCK_IS_ON));
     } /* End if */
     else
     {
	  options.CAPS_LOCK=OFF;
	  gtk_text_insert(GTK_TEXT(foo), NULL, NULL, NULL,
			  CAPS_LOCK_IS_OFF, strlen(CAPS_LOCK_IS_OFF));
     } /* End else */
} /* End capslock_toggle */

void shift_on(GtkWidget *w, gpointer data)
{
     /* Turn shift on */
     options.SHIFT=ON;

     gtk_text_insert(GTK_TEXT((GtkWidget *)data), NULL, NULL, NULL,
                     SHIFT_IS_ON, strlen(SHIFT_IS_ON));
} /* End shift_on */
