/* EXT_PROGS.C
 * For use with GTKeyboard by David Allen s2mdalle@titan.vcu.edu
 * 
 * This file is released under the terms of the GNU General Public License
 * see COPYING for details
 *
 * This module deals with non-gtkeyboard programs that can be called from 
 * within GTKeyboard.  Start out with simple network stuff.
 */

/* Includes */
#include "master.h"
#include "ext_progs.h"
#include "makeshift.h"

#ifdef MEM_DEBUGGING
#include "mem/memdebug.h"
#endif /* MEM_DEBUGGING */

void EXT_callprog(GtkWidget *w, int PROGTYPE)
{
     char *cmdline = NULL;
     char *args[4];

#ifdef SIGCHLD
     signal(SIGCHLD, SIG_IGN);
#endif /* SIGCHLD */
     
#ifdef SIGCLD
     signal(SIGCLD, SIG_IGN);
#endif /* SIGCLD */

     args[3] = NULL;

     if(!fork())
	  gtkeyboard_message(1,"Executing command...\n");
     else
     {
	  args[0] = g_new0_(char, 4);
	  strcpy(args[0],"-e");
	  
	  cmdline = g_new0_(char, strlen(RXVT_LOC) + 1);
	  strcpy(cmdline,RXVT_LOC);

	  /* This is where we check on which program they're actually 
	   * desirous of running.  
	   */
	  switch(PROGTYPE)
	  {
	  case PROG_TELNET:
	       args[1] = g_new_(char, strlen(TELNET_LOC) +1 );
	       strcpy(args[1],TELNET_LOC);
	       args[2] = NULL;
	       break;
	  case PROG_FTP:
	       args[1] = g_new_(char, strlen(FTP_LOC)+1);
	       strcpy(args[1],FTP_LOC);
	       args[2] = NULL;
	       break;
	  case PROG_LYNX:
	       args[1] = g_new_(char, strlen(LYNX_LOC) +1);
	       args[2] = g_new_(char, 50);
	       strcpy(args[1],LYNX_LOC);

	       /* Send people to the gtkeyboard page.  Yes it's blatant
		* self-promotion.  Cope.
		*/
	       strcpy(args[2],"http://opop.nols.com/gtkeyboard.html");
	       break;
	  default:
	       g_free_(args[0]);
	       args[0] = NULL;
	       args[1] = NULL;
	       args[2] = NULL;
	       break;
	  } /* End switch */

	  if(strlen(cmdline)>1)
	       execv(cmdline,args);   /* Actually execute the damn thing */
	  else
	  {
	       /* Keep this to stderr NOT Q */
	       gtkeyboard_error(4,"Can't execute!  Command line s trlen>!1\n",
				"Did you configure the program correctly?\n",
				"Check makeshift.h and make sure that the\n",
				"values point to valid executables.\n\n");
	       annoying_popup("Can't execute!");
	  }

	  /* Free up some memory so the user still has enough free mem
	   * to go look up their porn^H^H^H^H I mean run the program later
	   */
	  g_free_(cmdline);
	  g_free_(args[0]);
	  g_free_(args[1]);
	  g_free_(args[2]);
	  exit(0);     
     }
} /* End EXT_callprog */









