/* FILE_FOO.C
 * For use with GTKeyboard
 * written by David Allen s2mdalle@titan.vcu.edu
 * http://opop.nols.com/
 * 
 * This file is released under the terms of the GNU General Public License.
 *
 * This is mostly stuff to help deal with parsing the RCfile which we are
 * hopefully going to find in the user's home directory - something like
 * ~/.gtkeyboardrc and ~/.gtkeyboard-shortcuts
 *
 * Change:  9/5/99 g_strcasecmp is used instead of strcasecmp for 
 * portability.
 *
 */

/* Let the Makefile do this later if it wants to
 * #define DEBUGGING
 */

#define FILE_FOO_C
#include <stdio.h>
#include <stdlib.h>
#include "include/file_foo.h"
#include "master.h"

static char *prechomp                 (char *input);

#ifdef DEBUGGING
#define VERBOSEMSG(fmt, arg)    if(options.VERBOSE){fprintf(Q,fmt,arg);\
                                                    fflush(Q);}
#else /* not defined(DEBUGGING) */
#define VERBOSEMSG(fmt, arg)    ;
#endif

/* Read one line from the specified file pointer and return it as
 * as a pointer to char and so on.
 */
void FILE_readline(FILE *fp, char *buffer, const int maxlen)
{
     int index = 0 ;
     int x;
     char tmp;

     do
     {
          x = fread(&tmp,sizeof(char), 1, fp);
          if((x==0) || (tmp == '\n'))
	       buffer[index] = '\0'; /* Terminate the string with a NULL */
          else
	       buffer[index++] = tmp; /*  Add the character */
	  if(!(index<maxlen))
	  {
	       fprintf(Q,"Error on FILE_readline: index >= maxlen\n");
	       fflush(Q);
	       return;
	  }
     } while ((x != 0) && (tmp !='\n'));
     return;
} /* End FILE_readline */

char *expand_relative_pathname(char *filename)
{
     char *file = NULL;

     if(!filename)
     {
	  file = g_strdup_(options.extrafiles);
	  return(file);
     } /* End if */

     file = g_new0_(char, strlen(filename) + 
		    strlen(options.extrafiles) + 2);
     sprintf(file,"%s%s",options.extrafiles, filename);
     return(file);
} /* End expand_relative_pathname */

/* Prechomp:  Eliminates leading whitespace in a string.  Similar to 
 * the Perl5 regexp $foo =~ s/^\s*?//g;  I don't want to have to require
 * the regex library for this small task, so I'll do it myself.
 */
static char *prechomp(char *input)
{
     char *pc = input ;
     for(pc = input; *pc != '\0'; pc++)
     {
	  if (!isspace(*pc)) break;
     }
     return(pc);
} /* End prechomp */

int toolbar_parse(char *option, char *value)
{
     int val = OFF;
     
     fprintf(Q,"Option = %s value = %s\n",option,value);
     fflush(Q);
     if(option == (char *)NULL || value == (char *)NULL)
     {
	  fprintf(Q,"Toolbar:  No good args - returning.\n");
	  return(0);
     } /* End if */

     if(OPT_IS_GOOD(value))
	  val = ON;          /* Otherwise off */
     else
	  val = OFF;

     fprintf(Q,"Val is now %d.\n",val);
     fflush(Q);

     if(CHECK(option, "cut"))
	  Toolbar.TB_CUT = val;
     else if(CHECK(option, "copy"))
	  Toolbar.TB_COPY = val;
     else if(CHECK(option, "paste"))
	  Toolbar.TB_PASTE = val;
     else if(CHECK(option, "font"))
	  Toolbar.TB_FONT = val;
     else if(CHECK(option, "keyboard_font"))
	  Toolbar.TB_KEYBOARD_FONT = val;
     else if(CHECK(option,"window_grab"))
	  Toolbar.TB_WINDOW_GRAB = val;
     else if(CHECK(option,"raise"))
	  Toolbar.TB_RAISE = val;
     else if(CHECK(option,"lower"))
	  Toolbar.TB_LOWER = val;
     else if(CHECK(option,"window_ignore"))
	  Toolbar.TB_WINDOW_IGNORE = val;
     else if(CHECK(option, "open"))
	  Toolbar.TB_OPEN = val;
     else if(CHECK(option,"positioning"))
	  Toolbar.TB_POSITIONING = val;
     else if(CHECK(option, "save"))
	  Toolbar.TB_SAVE = val;
     else if(CHECK(option, "color"))
	  Toolbar.TB_COLOR = val;
     else if(CHECK(option,"new"))
	  Toolbar.TB_NEW = val;
     else if(CHECK(option, "reset_style"))
	  Toolbar.TB_RESET_STYLE = val;
     else if(CHECK(option, "status_report"))
	  Toolbar.TB_STATUS_REPORT = val;
     else if(CHECK(option, "shortcuts"))
	  Toolbar.TB_SHORTCUTS = val;
     else
     {
	  fprintf(Q,"Invalid argment in rcfile:  \"%s\"\n",option);
	  fflush(Q);
     } /* End else */
     return(1);
} /* End toolbar_parse */

/* Do the work in parsing the rcfile.  This is not a very good way to do 
 * this, but I honestly don't know any better, and nobody has stepped forward
 * to say, "Hey!  That's a really BRAINDAMAGED way to do it!  Try it this 
 * way..." so here it is.
 */
int parse_rcfile(FILE *fp, char *filename)
{
     char line[MAX_LINE_LENGTH];
     char linecopy[MAX_LINE_LENGTH];
     int val;
     unsigned int bit = 0;
     static char *delimiters = " :\t\r\n";
     char token_0[TOKEN_SIZE];
     char token_1[TOKEN_SIZE];
     char token_2[TOKEN_SIZE];
#ifndef PROD
     int old_verbose = options.VERBOSE;  /* So we can restore it later */

     options.VERBOSE = ON;
#endif /* PROD */

     while(!feof(fp))
     {
	  FILE_readline(fp, line, MAX_LINE_LENGTH);
	  if (*line == '#')
	       continue;
	  
	  /* Eliminate leading whitespace from string 
	   * The resulting length of prechomp is guaranteed to be
	   * less than or equal to line.  
	   */
	  strcpy(line, prechomp(line));
	  strcpy(linecopy, line);
	  
	  if(strlen(line)<3)
	  {
	       val = -666;       /* Whatever */
	       continue;
	  } /* End if */
	  else
	  {
	       char *ptoken_0 = strtok(line, delimiters);
	       char *ptoken_1 = (char *) NULL;
	       char *ptoken_2 = (char *) NULL;

	       if ((char *)NULL == ptoken_0 )
		    continue ;	/* empty line */

	       if ((char *)NULL != ptoken_0  &&  *ptoken_0 == '#' )
		    continue ;	/* comment line	*/
	       ptoken_1 = strtok(NULL, delimiters);
	       if ((char *)NULL != ptoken_1 )
		    ptoken_2 = strtok(NULL, delimiters);
	       strcpy(token_0 , ptoken_0 );
	       if ((char *)NULL != ptoken_1)
		    strcpy(token_1 , ptoken_1 );
	       
	       val = 0 ;
	       
	       if ((char *)NULL != ptoken_2)
	       {
		    strcpy(token_2, ptoken_2);
		    val = atoi(token_2);
		    
		    if(OPT_IS_GOOD(token_2))
			 val = ON;

		    /* Make sure that val is either 1 or 0 for options
		     * that are 1 bit
		     */
		    if(val) val = ON;
		    else    val = OFF;
	       } /* End if */
	       
	       if(CHECK(ptoken_0, "show") || CHECK(ptoken_0, "hide"))
	       {
		    if(CHECK(ptoken_0,"show"))
			 bit = 1;
		    else bit = 0;

		    if(CHECK(ptoken_1, "toolbar"))
			 ELEMENTS.toolbar = bit;
		    else if(CHECK(ptoken_1, "text"))
			 ELEMENTS.text = bit;
		    else if(CHECK(ptoken_1, "status"))
			 ELEMENTS.status = bit;
		    else if(CHECK(ptoken_1, "keyboard"))
			 ELEMENTS.keyboard = bit;
		    else if(CHECK(ptoken_1, "menubar"))
			 ELEMENTS.menubar = bit;
		    else if(CHECK(ptoken_1, "buttonbar"))
			 ELEMENTS.buttonbar = bit;
		    else 
		    {
			 fprintf(stderr,"Error:  unknown element \"%s\"\n",
				 ptoken_1); fflush(stderr);
		    } /* End else */
	       } /* End if */
	       else if(CHECK(ptoken_0, "toolbar"))
	       {
		    toolbar_parse(ptoken_1, ptoken_2);
		    continue;
	       } /* End if */
	       else if(CHECK(ptoken_0, "set"))
	       {
		    if(CHECK(ptoken_1, OPT_WORD_WRAP))
		    {
			 VERBOSEMSG("Set wordwrap to %d\n",val);
			 options.WORD_WRAP = val;
		    } /* End if */
		    else if(CHECK(ptoken_1, OPT_XPOS))
		    {
			 options.xpos = atoi(ptoken_2);
			 VERBOSEMSG("Set XPOS to %d.\n",options.xpos);
		    } /* End else if */
		    else if(CHECK(ptoken_1, OPT_YPOS))
		    {
			 options.ypos = atoi(ptoken_2);
			 VERBOSEMSG("Set YPOS to %d.\n",options.ypos);
		    } /* End else if */
		    else if(CHECK(ptoken_1, OPT_ASK_REMAP_ON_EXIT))
		    {
			 VERBOSEMSG("Set ASK_REMAP_ON_EXIT to %d\n",val);
			 options.ASK_REMAP_ON_EXIT = val;
		    } /* End else if */
		    else if(CHECK(ptoken_1, OPT_BORDERLESS))
		    {
			 VERBOSEMSG("Set BORDERLESS to %d\n",val);
			 options.BORDERLESS = val;
		    } /* End else if */
		    else if(CHECK(ptoken_1, OPT_INFO_POPUPS))
		    {
			 VERBOSEMSG("Set INFO_POPUPS to %d\n",val);
			 options.INFO_POPUPS = val;
		    } /* End else if */
		    else if(CHECK(ptoken_1, OPT_REDIRECT_POLICY))
		    {
			 if(CHECK(ptoken_2,"implicit"))
			 {
			      VERBOSEMSG("Setting redirect policy to %s\n",
					 ptoken_2);
			      options.REDIRECT_POLICY_IMPLICIT = ON;
			 }
			 else if(CHECK(ptoken_2,"explicit"))
			 {
			      VERBOSEMSG("Setting redirect policy to %s\n",
					 ptoken_2);
			      options.REDIRECT_POLICY_IMPLICIT = OFF;
			 }
			 else
			 {
			      fprintf(stderr,"Unknown redirect policy.\n");
			      fflush(stderr);
			 }    
		    } /* End else if */
		    else if(CHECK(ptoken_1,OPT_BUMP_AMOUNT))
		    {
			 VERBOSEMSG("Set BUMP_AMOUNT to %d\n",val);
			 GUI.BUMP_AMOUNT = atoi(ptoken_2);

			 /* Use the absolute value */
			 if(GUI.BUMP_AMOUNT < 0)
			      GUI.BUMP_AMOUNT *= -1;
		    } /* End else if */
		    else if(CHECK(ptoken_1,OPT_ASK_SAVE))
		    {
			 VERBOSEMSG("Set ASK_SAVE_ON_EXIT to %d\n",val);
			 options.ASK_SAVE_ON_EXIT = val;
		    } /* End else if */
		    else if(CHECK(ptoken_1, OPT_ALL_CAPS))
			 options.CAPS_LOCK = val;
		    /* EYECANDY is a synonym for RANDOMIZE_XPMS */
		    else if(CHECK(ptoken_1, OPT_RANDOMIZE_XPMS) ||
			    CHECK(ptoken_1, OPT_EYECANDY))
		    {
			 VERBOSEMSG("Set RANDOMIZE_XPMS to %d\n",val);
			 options.RANDOMIZE_XPMS = val;
		    } /* End else if */
		    else if(CHECK(ptoken_1, OPT_STATUS_LOGGING))
		    {
			 VERBOSEMSG("Set STATUS_LOGGING to %d\n",val);
			 options.STATUS_LOGGING = val;
		    } /* End else */
		    else if(CHECK(ptoken_1, OPT_DISPLAY_TOOLBAR))
		    {
			 VERBOSEMSG("Set SHOW_TOOLBAR to %s\n",ptoken_2)
			 options.DISPLAY_TOOLBAR = val;
		    } /* End else if */
		    else if(CHECK(ptoken_1, OPT_EXTRA_FILES))
		    {
			 if(ptoken_2)
			 {
			      VERBOSEMSG("Set EXTRA_FILES to %s\n",ptoken_2);
			      CONDFREE(options.extrafiles);
			      options.extrafiles = g_strdup_(ptoken_2);
			 } /* End if */
			 else
			 {
			      VERBOSEMSG("Bad EXTRA_FILES statement: %s.\n",
					 "missing directory path");
			 } /* End else */
		    } /* End else if */
		    else if(CHECK(ptoken_1, OPT_SHOW_MENU))
		    {
			 VERBOSEMSG("Set SHOW_MENU to %s\n",ptoken_2);
			 options.SHOW_MENU = val;
		    }
		    else if(CHECK(ptoken_1, OPT_PULLOFFS))
		    {
			 if(g_strcasecmp(ptoken_2,"left") == 0)
			 {
			      GUI.PULLOFFS_SIDE = GTK_POS_LEFT;
			      VERBOSEMSG("Pulloffs set to %s side\n",ptoken_2);
			 } /* End if */
			 else if(g_strcasecmp(ptoken_2,"right") == 0)
			 {
			      GUI.PULLOFFS_SIDE = GTK_POS_RIGHT;
			      VERBOSEMSG("Pulloffs set to %s side\n",ptoken_2);
			 } /* End else if */
			 else if(g_strcasecmp(ptoken_2,"top") == 0)
			 {
			      GUI.PULLOFFS_SIDE = GTK_POS_TOP;
			      VERBOSEMSG("Pulloffs set to %s side\n",ptoken_2);
			 } /* End else if */
			 else if(g_strcasecmp(ptoken_2,"bottom") == 0)
			 {
			      GUI.PULLOFFS_SIDE = GTK_POS_BOTTOM;
			      VERBOSEMSG("Pulloffs set to %s side\n",ptoken_2);
			 } /* End else if */
			 else
			 {
			      fprintf(Q,"Invalid pulloffs value: %s\n",
				      ptoken_2);
			      fflush(Q);
			 } /* End else */
		    } /* End else if OPT_PULLOFFS */
		    else if(CHECK(ptoken_1, OPT_DISPLAY_ROW_BUTTONS))
		    {
			 VERBOSEMSG("Set DISPLAY_ROW_BUTTONS to %s\n",
				    ptoken_2);
			 options.DISPLAY_ROW_BUTTONS = val;
		    }
		    else if(CHECK(ptoken_1, OPT_WORKING_FILE))
		    {
			 VERBOSEMSG("Set WORKINGFILE to %s\n",ptoken_2);
			 options.workingfile = g_strdup_( ptoken_2 );
		    } /* End else if */
		    else if(CHECK(ptoken_1, OPT_SHOW_TEXT))
		    {
			 VERBOSEMSG("Set SHOW_TEXT to %s\n",ptoken_2);
			 options.SHOW_ALL_WINDOWS = ( val == 0 ) ? OFF : ON ;
		    }
		    else if(CHECK(ptoken_1, OPT_RAND_STRING_LENGTH))
		    {
			 options.RANDOM_STRING_CHARACTERS = atoi(ptoken_2);
			 if(options.RANDOM_STRING_CHARACTERS<0)
			      options.RANDOM_STRING_CHARACTERS*=-1;
			 
			 VERBOSEMSG("Set RANDOM_STRLEN to %d\n",
				    options.RANDOM_STRING_CHARACTERS);
		    } /* End else if */
		    else if(CHECK(ptoken_1, OPT_SEND_TO_BOTH_WINDOWS))
		    {
			 options.SEND_TO_BOTH_WINDOWS = val;
			 VERBOSEMSG("Set SEND_TO_BOTH_WINDOWS to %s\n",
				    ptoken_2);
		    } /* End else if */
		    else
		    {
			 fprintf(Q,"Line \"%s\": Eh?\n",linecopy);
			 fflush(Q);
		    } /* End else */
	       } /* End else if */
	       else   /* Unknown statmenet */
	       {
		    fprintf(Q,"\"%s\" :  ",linecopy);
		    fprintf(Q,"What are you talking about?\n");
		    continue;
	       } /* End else */
	  } /* End else*/
     } /* End while */
     
#ifndef PROD
     options.VERBOSE = old_verbose;    /* Restore original state of
					* options.VERBOSE
					*/
#endif /* PROD */

     /* Flush various streams */
     flush_everything();
     return(1);
} /* End parse_rcfile */

/* Loads the shortcuts into the program from SHORTCUTS_FILENAME */
int gtkeyboard_load_shortcuts(void)
{
     FILE *fp;
     char filename[1024];
     register int x=0;
     char line[SHORTCUT_MAX_LENGTH];

     sprintf(filename,"%s/%s",options.home, SHORTCUTS_FILENAME);

     if((fp = fopen(filename,"r")) == NULL)
     {
	  gtkeyboard_error(5,"Couldn't open ",filename," for reading:  ",
			   g_strerror(errno),"\n",
			   "A shortcuts file will be created for you.");
	  return(0);
     } /* End if */
     
     x=0;       /* Line number and array subscript for options.shortcuts[x] */
     
     while(!feof(fp) && x<SHORTCUTS)
     {
	  FILE_readline(fp, line, SHORTCUT_MAX_LENGTH);
	  if(g_strcasecmp(line,"NONE")==0)
	       options.shortcuts[x] = (char *)NULL;
	  else
	  {
	       if(line)
	       {
		    options.shortcuts[x] = g_strdup_(line);
	       }
	       else
		    options.shortcuts[x] = (char *)NULL;
	  }
	  
	  if(!line)
	  {
	       gtkeyboard_message(1,"Shortcut nonexistant from file.\n");
	  } /* End if */
	  x++;
     } /* End while */
     
     fclose(fp);     /* Close the file */
     return(1);
} /* End gtkeyboard_load_shortcuts */

/* Saves shortcuts to 
 * $HOME/SHORTCUTS_FILENAME
 */
void gtkeyboard_save_shortcuts(GtkWidget *widget, gpointer data)
{
     FILE *fp;
     char filename[1024];
     register int x=0;

     /* Destroy calling widget window if needed */
     if(data)
	  gtk_widget_destroy((GtkWidget *)data);

     /* This is begging for a segfault so let's not do it. */
     if((strlen(options.home) + strlen(SHORTCUTS_FILENAME) + 2) >=1024)
     {
	  gtkeyboard_error(2,"Can't save shortcuts:  filename overflow.\n",
			   "Please report this to the maintainer.\n");
	  return;
     } /* End if */

     sprintf(filename,"%s/%s",options.home, SHORTCUTS_FILENAME);
     fprintf(Q,"Saving shortcuts to \"%s\"\n",filename);

     if((fp = fopen(filename,"w")) == NULL)
     {
	  gtkeyboard_error(5,"Cannot open ", filename, " for writing:  ",
			   g_strerror(errno), "\nShortcuts NOT SAVED.\n");
	  return;
     } /* End if */

     /* Save each shortcut on one line.  Real simple.  Hopefully */
     for(x=0; x<SHORTCUTS; x++)
     {
	  if(!options.shortcuts[x] || options.shortcuts[x] == (char *)NULL)
	       fprintf(fp,"NONE\n");
	  else
	       fprintf(fp,"%s\n",options.shortcuts[x]);
     } /* End for */

     fclose(fp);

     gtkeyboard_message(2,"Shortcuts saved to file ",filename,"\n");
} /* End gtkeyboard_save_shortcuts */

/* Copies a default rcfile from somewhere to the input filename */
int setup_default_rcfile(char *file)
{
     FILE *fp;
     FILE *dflt;
     char buffer[1024];
     char buffer2[1024];
     
     if((fp = fopen(file,"w"))==NULL)
     {
	  fprintf(Q,"Couldn't open %s for writing - cannot create ",file);
	  fprintf(Q,"default .gtkeyboardrc!\n");
	  fflush(Q);
	  return(0);
     } /* End if */
     else
     {
	  /* Try to open the distribution-provided rcfile first */
	  sprintf(buffer,"%s/%s",options.extrafiles, PROVIDED_RCFILE);
	  
	  if((dflt = fopen(buffer,"r")) == NULL)
	  {
	       /* Ok, fine, so we can't open the default one we provided 
		* with the distribution.  We'll just give the user a really
		* short and crappy one.
		*/
	       fprintf(Q,"Couldn't open %s: %s.\n",buffer,g_strerror(errno));
	       fprintf(Q,"Fine then!  We'll try something else...\n");
	       fflush(Q);

	       fprintf(fp,"%s",DEFAULT_RCFILE);
	       fclose(fp);
	       fprintf(Q,"Success writing default .gtkeyboardrc\n");
	       fprintf(Q,"You can edit it to make changes later by editing ");
	       fprintf(Q,"%s\n",file);
	       fflush(Q);
	       return(1);
	  }
	  else
	  {
	       while(!feof(dflt))
	       {
		    FILE_readline(dflt, buffer2, 1024);   
		    /* Add a newline because FILE_readline kills them off */
		    fprintf(fp,"%s\n",buffer2);
	       } /* End while */
	       fflush(fp);
	       fclose(dflt);     
	       fclose(fp);

	       fprintf(Q,"Successfully wrote .gtkeyboardrc from ");
	       fprintf(Q,"default.  (Distribution provided)\n");
	  } /* End else */
     } /* End else */
     flush_everything();
     return(1);
} /* End setup_default_rcfile */
