/* File handling and querying
 *
 * Written by David Allen <s2mdalle@titan.vcu.edu>
 * http://opop.nols.com/
 * Released under the terms of the GNU General Public License
 */

#define FILE_UTILS_C
#include "master.h"

/* Returns a list of filenames that contain substr located in directory */
GList *get_files(char *directory, char *substr)
{
     struct dirent *dir_info;
     DIR *dir;
     GList *retty = NULL;
     char *dupptr;

     dir = opendir(directory);

     if(!dir)
     {
	  fprintf(Q,"Couldn't open directory %s:  %s\n",
		  directory, g_strerror(errno));
	  fflush(Q);
	  return(NULL);
     } /* End if */
     
     dir_info = readdir(dir);
     
     if(!dir_info)
     {
	  fprintf(Q,"Couldn't read directory %s: %s\n",
		  directory, g_strerror(errno)); 
	  fflush(Q);
	  return(NULL);
     }

     while(dir_info)
     {
	  if(strstr(substr, dir_info->d_name))
	  {
#ifdef DEBUGGING
	       fprintf(Q,"Appending \"%s\" which matches \"%s\"\n",
		       dir_info->d_name, substr); 
	       fflush(Q);
#endif
	       dupptr = g_strdup_(dir_info->d_name);
	       retty = g_list_append(retty, dupptr);
	  } /* End if */
     } /* End while */

     closedir(dir);
     return(retty);
} /* End get_files */

int is_regular_file(const char *filename)
{
     struct stat file_info;
     int x;

     STAT_AND_RETURN_IF_BAD(filename, x, file_info);

     return(S_ISREG(file_info.st_mode));
} /* End is_regular_file() */

int is_fifo(const char *filename)
{
     struct stat file_info;
     int x;

     STAT_AND_RETURN_IF_BAD(filename, x, file_info);

     return(S_ISFIFO(file_info.st_mode));
} /* End is_fifo() */

int is_link(const char *filename)
{
     struct stat file_info;
     int x;

     STAT_AND_RETURN_IF_BAD(filename, x, file_info);

     return(S_ISLNK(file_info.st_mode));
} /* End is_link() */

int is_executable(const char *filename)
{
     struct stat file_info;
     int x;

     STAT_AND_RETURN_IF_BAD(filename, x, file_info);
     
     return((file_info.st_mode & S_IXUSR) ||
	    (file_info.st_mode & S_IXGRP) ||
	    (file_info.st_mode & S_IXOTH));
} /* End is_executable */

int file_exists(const char *filename)
{
     struct stat file_info;
     int x;

     STAT_AND_RETURN_IF_BAD(filename, x, file_info);

     return(1);
} /* End file_exists */
