/* UI.h
 *
 * This file is for use with GTKeyboard, and is released under the terms of
 * the GNU General Public License
 *
 * Written by David Allen <s2mdalle@titan.vcu.edu>
 * http://opop.nols.com/
 *
 * Prototypes for ui.c
 */

#define UI_H
#ifndef MASTER_H
#  include "../master.h"
#endif /* MASTER_H */

#ifndef HTML_H
#  include "html.h"
#endif /* HTML_H */

/* Just make these unique. */
#define ORIGIN                              20
#define BR_CORNER                           21
#define UR_CORNER                           22
#define BL_CORNER                           23
#define UL_CORNER                           24
#define WIN_CENTER                          25
#define T_CENTER                            26
#define B_CENTER                            27
#define L_CENTER                            28
#define R_CENTER                            29
#define NO_HANDLER                          666

#define DIMENSIONS_ERROR(winfu)             (winfu.x==-1 && winfu.y==-1 && \
                                           winfu.width==-1 && winfu.height==-1)
/* Prototypes for ui.c */
GtkWidget *create_toolbar                    (GtkWidget *emitter, 
					      gpointer data);
void EasterEgg                               (GtkWidget *emitter, 
					      gpointer data);
void setup_output_box                        (void);
void CreateMainWindow                        (void);
void select_all                              (GtkWidget *w, gpointer data);
void make_item_factory                       (void);
void move_gtkeyboard                         (const int x, const int y);
GtkWidget *make_browse_button                (GtkWidget *insert_into);
void browse                                  (GtkWidget *emitter, 
					      gpointer data);
void move_window                             (GtkWidget *emitter, 
					      int POSITION);
window_dimensions gtkeyboard_query_window    (Window w);
void gtkeyboard_print_coords                 (GtkWidget *emitter, 
					      gpointer data);
void redirect_explicit                       (GtkWidget *emitter, 
					      gpointer data);
void redirect_implicit                       (GtkWidget *emitter, 
					      gpointer data);
void move_gtkeyboard_offset                  (const int x, const int y);
void gtkeyboard_raise                        (void);
void gtkeyboard_lower                        (void);
void gtkeyboard_hide                         (void);
void gtkeyboard_show                         (void);
void set_handle_box_opts                     (GtkWidget *handlebox);

#ifdef UI_C
GtkItemFactoryEntry menu[] =
{
     /* FILE MENU */
     {"/_File",            NULL,          0,         0,  "<Branch>"   },
     {"/_File/tearoff1",   NULL,    Showmenu, (gint)FILEM, "<Tearoff>"},
     {"/_File/_New",       "<control>N",  NewFile,                 0  },
     {"/_File/_Open",      "<control>O",  OpenFile,                0  },
     {"/_File/_Save",      "<control>S",  save_output_text,        0  },
     {"/_File/Save _As",   "<control>A",  save_file_as,            0  },
     {"/_File/_Quit",      "<control>Q",  (void *)final_check,     0  },

     /*
      * This option is no longer included because it would cause mem probs
      *{"/_File/FastExit",        NULL,  (void *)byebye,   ((gint)0)  },
     */

     /* EDIT MENU */
     {"/_Edit",            NULL,          0,         0,  "<Branch>"   },
     {"/_Edit/tearoff1",   NULL,    Showmenu, (gint)EDITM, "<Tearoff>"},
     {"/_Edit/_Cut",       "<control>X",  CutText,         0,      0  },
     {"/_Edit/_Copy",      "<control>C",  CopyText,        0,      0  },
     {"/_Edit/_Paste",     "<control>V",  PasteText,       0,      0  },
#if 0
     {"/_Edit/Find",       "<control>F",  find_dialog,     0,      0  },
#endif

#ifndef INCLUDE_HTML
#  define INCLUDE_HTML
#endif /* INCLUDE_HTML */

#ifdef INCLUDE_HTML
     /* HTML MENU */
     {"/_HTML",                   NULL,        0,         0,  "<Branch>" },
     {"/_HTML/tearoff1",          NULL,Showmenu, (gint)HTMLM, "<Tearoff>"},
     {"/_HTML/Document",          NULL,          0,         0,"<Branch>" },
     {"/_HTML/Document/Doctype",  NULL,          html_doctype,        0  },
     {"/_HTML/Document/Open Doc", NULL,          html_html,           0  },
     {"/_HTML/Document/Close Doc",NULL,          html_html_close,     0  },
     {"/_HTML/Document/Head",     NULL,          html_head,           0  },
     {"/_HTML/Document/Close Head",NULL,         html_head_close,     0  },
     {"/_HTML/Document/Title",    NULL,          html_title,          0  },
     {"/_HTML/Body",              NULL,          0,       0,  "<Branch>" },
     {"/_HTML/Body/Open Body",    NULL,          html_body,           0  },
     {"/_HTML/Body/Close Body",   NULL,          html_body_close,     0  },
     {"/_HTML/Text",              NULL,          0,       0,  "<Branch>" },
     {"/_HTML/Text/Bold",         NULL,          html_bold,           0  },
     {"/_HTML/Text/Strong",       NULL,          html_strong,         0  },
     {"/_HTML/Text/Italics",      NULL,          html_italics,        0  },
     {"/_HTML/Text/Center",       NULL,          html_center,         0  },
     {"/_HTML/Text/Preformatted", NULL,          html_pre,            0  },
     {"/_HTML/Text/Font Color",   NULL,          html_font_color,     0  },
     {"/_HTML/Text/Font Size",    NULL,          html_font_size,      0  },
     {"/_HTML/Text/Strikeout",    NULL,          html_strike,         0  },
     {"/_HTML/Text/Subscript",    NULL,          html_sub,            0  },
     {"/_HTML/Text/Superscript",  NULL,          html_super,          0  },
     {"/_HTML/Text/Space",        NULL,          html_nbsp,           0  },
     {"/_HTML/Text/Blink (evil)", NULL,          html_blink,          0  },
     {"/_HTML/Headers",           NULL,          0,         0,  "<Branch>" },
     {"/_HTML/Headers/H1",        NULL,          html_header,    (int)H1 },
     {"/_HTML/Headers/H2",        NULL,          html_header,    (int)H2 },
     {"/_HTML/Headers/H3",        NULL,          html_header,    (int)H3 },
     {"/_HTML/Headers/H4",        NULL,          html_header,    (int)H4 },
     {"/_HTML/Headers/H5",        NULL,          html_header,    (int)H5 },
     {"/_HTML/Headers/H6",        NULL,          html_header,    (int)H6 },
     {"/_HTML/Lists",             NULL,          0,          0,  "<Branch>" },
     {"/_HTML/Lists/New List",    NULL,          html_list,           0  },
     {"/_HTML/Lists/List Item",   NULL,          html_list_item,      0  },
     {"/_HTML/Tables",            NULL,          0,          0,  "<Branch>"},
     {"/_HTML/Tables/New Table",  NULL,          html_table,          0  },
     {"/_HTML/Tables/Table Row",  NULL,          html_table_row,      0  },
     {"/_HTML/Tables/Table Data", NULL,          html_table_data,     0  },
     {"/_HTML/Protocols",         NULL,          0,          0,  "<Branch>"},
     {"/_HTML/Protocols/HTTP",    NULL,          html_http,           0  },
     {"/_HTML/Protocols/FTP",     NULL,          html_ftp,            0  },
     {"/_HTML/Protocols/News",    NULL,          html_news,           0  },
     {"/_HTML/Protocols/Telnet",  NULL,          html_telnet,         0  },
     {"/_HTML/Protocols/Gopher",  NULL,          html_gopher,         0  },
     {"/_HTML/Protocols/WWW",     NULL,          html_www,            0  },
     {"/_HTML/Internal",          NULL,          0,        0,  "<Branch>"},
     {"/_HTML/Internal/BGCOLOR",  NULL,  html_internal,(gint)HTML_BGCOLOR},
     {"/_HTML/Internal/TEXT",     NULL,  html_internal,   (gint)HTML_TEXT},
     {"/_HTML/Internal/LINK",     NULL,  html_internal,   (gint)HTML_LINK},
     {"/_HTML/Misc",              NULL,          0,          0,"<Branch>"},
     {"/_HTML/Misc/SSI",          NULL,          html_ssi,            0  },
     {"/_HTML/Misc/HR",           NULL,          html_hr,             0  },
     {"/_HTML/Misc/Linebreak",    NULL,          html_linebreak,      0  },
     {"/_HTML/Misc/mailto",       NULL,          html_mailto,         0  },
     {"/_HTML/Misc/Comment",      NULL,          html_comment,        0  },
     {"/_HTML/Misc/Close Unnamed",NULL,          html_close_unnamed,  0  },
     {"/HTML/ ",                  NULL,      0,       0,  "<Separator>"  },
     {"/_HTML/Paragraph",         NULL,          html_paragraph,      0  },
     {"/_HTML/Image",             NULL,          html_image,          0  },
     {"/_HTML/Link",              NULL,          html_link,           0  },
     {"/_HTML/Full Template",     NULL,          html_template,       0  },
#endif /* INCLUDE_HTML */

     /* SHORTCUTS MENU */
     {"/_Shortcuts",              NULL,          0,         0, "<Branch>" },
     {"/_Shortcuts/tearoff1",     NULL, Showmenu,(gint)SHORTCUTSM,"<Tearoff>"},
     {"/_Shortcuts/UserDef", NULL,          0,           0, "<Branch>"},
     {"/_Shortcuts/UserDef/Define", NULL,    define_shortcuts,        0   },
     {"/_Shortcuts/UserDef/#1",     NULL,         do_shortcut,   (int)0   },   
     {"/_Shortcuts/UserDef/#2",     NULL,         do_shortcut,   (int)1   },   
     {"/_Shortcuts/UserDef/#3",     NULL,         do_shortcut,   (int)2   },   
     {"/_Shortcuts/UserDef/#4",     NULL,         do_shortcut,   (int)3   },   
     {"/_Shortcuts/UserDef/#5",     NULL,         do_shortcut,   (int)4   },   
     {"/_Shortcuts/UserDef/#6",     NULL,         do_shortcut,   (int)5   },   
     {"/_Shortcuts/External Progs", NULL,          0,         0, "<Branch>"},
     {"/_Shortcuts/External Progs/Telnet",NULL,EXT_callprog,(int)PROG_TELNET},
     {"/_Shortcuts/External Progs/FTP", NULL,  EXT_callprog,(int)PROG_FTP},
     {"/_Shortcuts/External Progs/Lynx", NULL, EXT_callprog,(int)PROG_LYNX},
     {"/_Shortcuts/External Progs/(More later)", NULL, NULL},
     {"/_Shortcuts/Quoting",      NULL,          0,         0, "<Branch>"},
     {"/_Shortcuts/Quoting/Quote",NULL,          SC_quote,            0   },
     {"/_Shortcuts/Quoting/Single Quote", NULL,  SC_single_quote,     0   },
     {"/_Shortcuts/Quoting/Book Quote",   NULL,  SC_book_quote,       0   },
     {"/_Shortcuts/Quoting/Back Quote",   NULL,  SC_backquote,        0   },
     {"/_Shortcuts/Quoting/Paragraph",    NULL,  SC_paragraph,        0   },
     {"/_Shortcuts/Quoting/Surround",     NULL,  SC_surround,         0   },
     {"/_Shortcuts/Group",        NULL,          0,         0, "<Branch>" },
     {"/_Shortcuts/Group/Parentheses",  NULL,    SC_parens,           0   },
     {"/_Shortcuts/Group/Angle Bracket", NULL,   SC_angle,            0   },
     {"/_Shortcuts/Group/Dbl Angle",    NULL,    SC_double_angle,     0   },
     {"/_Shortcuts/Group/Hi-Hat",       NULL,    SC_hihat,            0   },
     {"/_Shortcuts/Emoticons",    NULL,          0,         0, "<Branch>" },
     {"/_Shortcuts/Emoticons/Smiley",     NULL,  SC_smiley,           0   },
     {"/_Shortcuts/Emoticons/Winky",      NULL,  SC_winky,            0   },
     {"/_Shortcuts/Emoticons/Rose",       NULL,  SC_rose,             0   },
     {"/_Shortcuts/ ",            NULL,      0,  0,        "<Separator>"  },
     {"/_Shortcuts/Character Count",      NULL,  SC_report_length,    0   },
     {"/_Shortcuts/Delete Line",          NULL,  delete_line,         0   },
     {"/_Shortcuts/Pseudo-Underline", NULL,      SC_pseudo_underline, 0   },
     {"/_Shortcuts/Emphasize",    NULL,          SC_emphasize,        0   },
     {"/_Shortcuts/Random String",        NULL,  SC_random_string,    0   },

     /* OPTIONS MENU */
     {"/Options",                   NULL,       0,         0,     "<Branch>" },
     {"/Options/tearoff1",          NULL,Showmenu,(gint)OPTIONSM, "<Tearoff>"},

     {"/Options/Move",              NULL,       0,         0,     "<Branch>" },
#ifdef USE_REMOTE_CONTROL
     {"/Options/Move/Remote Control",        NULL,gtkeyboard_remote_control,0},
#endif /* USE_REMOTE_CONTROL */
     {"/Options/Move/ ",                 NULL,   0,       0,   "<Separator>" },
     {"/Options/Move/to Origin",             NULL,move_window,   (int)ORIGIN },
     {"/Options/Move/Center",                NULL,move_window,(int)WIN_CENTER},
     {"/Options/Move/To bottom right corner",NULL,move_window,(int)BR_CORNER },
     {"/Options/Move/To bottom left corner", NULL,move_window,(int)BL_CORNER },
     {"/Options/Move/To upper right corner", NULL,move_window,(int)UR_CORNER },
     {"/Options/Move/To upper left corner",  NULL,move_window,(int)UL_CORNER },
     {"/Options/Move/Top centered",          NULL,move_window, (int)T_CENTER },
     {"/Options/Move/Bottom centered",       NULL,move_window, (int)B_CENTER },
     {"/Options/Move/Right centered",        NULL,move_window, (int)R_CENTER },
     {"/Options/Move/Left centered",         NULL,move_window, (int)L_CENTER },

     {"/Options/Looks",             NULL,        0,      0,       "<Branch>" },
     {"/Options/Looks/Colors",      NULL,        0,      0,       "<Branch>" },
     {"/Options/Looks/Colors/Set Colors", NULL, color_button_clicked,     0  },
     {"/Options/Looks/Colors/Reset Style",NULL,reset_application_style,  0   },
     {"/Options/Looks/Fonts/Set Keyboard Font",NULL,KeyboardFontSelect,   0  },
     {"/Options/Looks/Fonts/Set Text Font",    NULL, FontSelect,          0  },
     {"/Options/Looks/Text",        NULL,        0,      0,       "<Branch>" },
     {"/Options/Looks/Text/Hide",   NULL,     hide_widget, (gint)HIDE_OUTPUT },
     {"/Options/Looks/Text/Show",   NULL,   unhide_widget, (gint)SHOW_OUTPUT },
     {"/Options/Looks/Menu",        NULL,        0,      0,       "<Branch>" },
     {"/Options/Looks/Menu/Hide",   NULL,     hide_widget,  (gint)HIDE_MENUS },
     {"/Options/Looks/Menu/Show",   NULL,   unhide_widget,  (gint)SHOW_MENUS },
     {"/Options/Looks/Status",      NULL,        0,      0,       "<Branch>" },
     {"/Options/Looks/Status/Hide", NULL,     hide_widget, (gint)HIDE_STATUS },
     {"/Options/Looks/Status/Show", NULL,   unhide_widget, (gint)SHOW_STATUS },
     {"/Options/Looks/Keyboard",    NULL,        0,      0,       "<Branch>" },
     {"/Options/Looks/Keyboard/Hide", NULL,   hide_widget,(gint)HIDE_KEYBOARD},
     {"/Options/Looks/Keyboard/Show", NULL, unhide_widget,(gint)SHOW_KEYBOARD},
     {"/Options/Looks/Toolbar",     NULL,        0,        0,     "<Branch>" },
     {"/Options/Looks/Toolbar/Hide",  NULL,   hide_widget,(gint)HIDE_TOOLBAR },
     {"/Options/Looks/Toolbar/Show",  NULL, unhide_widget,(gint)SHOW_TOOLBAR },
     {"/Options/Looks/Buttonrow",     NULL,        0,      0,     "<Branch>" },
     {"/Options/Looks/Buttonrow/Hide", NULL,   hide_widget,
      (gint)HIDE_BOTTOM_ROW_BUTTONS },
     {"/Options/Looks/Buttonrow/Show", NULL,  unhide_widget,
      (gint)SHOW_BOTTOM_ROW_BUTTONS },

#ifdef USE_GDK_CURSORS
     {"/Options/Looks/Cursor",    NULL,           0,          0,  "<Branch>" },
     {"/Options/Looks/Cursor/X",  NULL,    switch_cursor, (gint)GDK_X_CURSOR },
     {"/Options/Looks/Cursor/Arrow",NULL,  switch_cursor,    (gint)GDK_ARROW },
     {"/Options/Looks/Cursor/Boat", NULL,  switch_cursor,     (gint)GDK_BOAT },
     {"/Options/Looks/Cursor/Bogosity",NULL,switch_cursor,(gint)GDK_BOGOSITY },
     {"/Options/Looks/Cursor/Clock",NULL,  switch_cursor,    (gint)GDK_CLOCK },
     {"/Options/Looks/Cursor/Cross",NULL,  switch_cursor,    (gint)GDK_CROSS },
     {"/Options/Looks/Cursor/Gobbler",NULL,switch_cursor,  (gint)GDK_GOBBLER },
     {"/Options/Looks/Cursor/Heart", NULL, switch_cursor,    (gint)GDK_HEART },
     {"/Options/Looks/Cursor/Hand",  NULL, switch_cursor,    (gint)GDK_HAND1 },
     {"/Options/Looks/Cursor/Pencil", NULL,switch_cursor,   (gint)GDK_PENCIL },
     {"/Options/Looks/Cursor/Shuttle",NULL,switch_cursor,  (gint)GDK_SHUTTLE },
     {"/Options/Looks/Cursor/Spider", NULL,switch_cursor,   (gint)GDK_SPIDER },
     {"/Options/Looks/Cursor/Trek",  NULL, switch_cursor,     (gint)GDK_TREK },
     {"/Options/Looks/Cursor/Target",NULL, switch_cursor,   (gint)GDK_TARGET },
     {"/Options/Looks/Cursor/Default",NULL,switch_cursor, (gint)GDK_X_CURSOR },
#endif /* USE_GDK_CURSORS */

     {"/Options/Looks/Hide All but Keyboard",NULL,   HideAllButKeyboard,   0 },
     {"/Options/Looks/Show All",             NULL,   ShowAll,              0 },
     {"/Options/Looks/Raise Window",         NULL,   gtkeyboard_raise,     0 },
     {"/Options/Looks/Lower Window",         NULL,   gtkeyboard_lower,     0 },

     {"/Options/Output",    NULL,          0,        0,          "<Branch>"  },
     {"/Options/Output/Clear Status",  NULL,clear_wrapper,(gint)CLEAR_STATUS },
     {"/Options/Output/Clear All Text",NULL,clear_wrapper,(gint)CLEAR_TEXT   },
     {"/Options/Output/Choose A Window", NULL,      grab_a_window,        0  },
     {"/Options/Output/Ignore Window",   NULL,      ignore_window,        0  },
     {"/Options/Output/System Console",  NULL, change_output, OUTPUT_CONSOLE },
     {"/Options/Output/Printer",         NULL, change_output, OUTPUT_PRINTER },
     {"/Options/Output/Temp file",       NULL,change_output, OUTPUT_TEMPFILE },

     {"/Options/Redirection", NULL,        0,        0,           "<Branch>" },
     {"/Options/Redirection/Explicit",   NULL, redirect_explicit,          0 },
     {"/Options/Redirection/Implicit",   NULL, redirect_implicit,          0 },
     
     {"/Options/ ",                      NULL,   0,       0,  "<Separator>"  },

     /* The redirect/ignore options are important so include them twice for
      * ease of location.
      */
     {"/Options/Choose a Layout File", "<control>L",
                                           custom_keyboard_file_dialog,   0  },
     {"/Options/Redirect to a Window",   NULL,  grab_a_window,            0  },
     {"/Options/Ignore a Window",        NULL,  ignore_window,            0  },
     {"/Options/Define Shortcuts",       NULL,  define_shortcuts,         0  },
     {"/Options/Toggle Word Wrap",       NULL,  toggle_wordwrap,          0  },
     {"/Options/Toggle Save on Exit",    NULL,  toggle_ask_save_on_exit,  0  },
     {"/Options/Toggle Status Logging",  NULL,  status_logging_toggle,    0  },
     {"/Options/Toggle Double Key Sending", NULL,toggle_double_data,      0  },
     {"/Options/Reset Looks",            NULL,  reset_application_style,  0  },
#ifdef USE_REMOTE_CONTROL
     {"/_Options/Remote Control",        NULL,  gtkeyboard_remote_control,0  },
#endif /* USE_REMOTE_CONTROL */

     /* MISC MENU */
     {"/_Misc",                   NULL,          0,         0,  "<Branch>"   },
     {"/_Misc/tearoff1",          NULL,   Showmenu, (gint)MISCM, "<Tearoff>" },
     {"/_Misc/Timestamp",         NULL,          SC_timestamp,         0  },
     {"/_Misc/Status Report",     NULL,          status_report,        0  },
     {"/_Misc/Beginning of line", NULL,          (void *)goto_bol,     0  },
     {"/_Misc/End of line",       NULL,          (void *)goto_eol,     0  },
     {"/_Misc/Beginning of file", NULL,          (void *)goto_bof,     0  },
     {"/_Misc/End of file",       NULL,          (void *)goto_eof,     0  },
     {"/_Misc/Show Coordinates",  NULL,       gtkeyboard_print_coords, 0  },
     {"/_Misc/Remap keyboard",    NULL,       remap_keyboard_really,   0  },
     
     /* Help menu */
     {"/_Help",                   NULL,    0,         0,    "<LastBranch>"},
     {"/_Help/tearoff1",          NULL,    Showmenu, (gint)HELPM, "<Tearoff>"},
     {"/_Help/_About",            NULL,    program_info,(gint)INFO_ABOUT  },
     {"/_Help/License",           NULL,    program_info,(gint)INFO_LICENSE},
     {"/_Help/What's New",        NULL,    program_info,(gint)INFO_CHANGELOG},
     {"/_Help/Splash",            NULL,    splash,                     0  },
}; /* End of menu (ItemFactory declaration) */

static int nmenu_items = sizeof(menu) / sizeof (menu[0]);

#endif /* UI_C */



