/* KMANIP.C
 * Written by David Allen <s2mdalle@titan.vcu.edu>
 * http://opop.nols.com/
 *
 * This file is distributed under the terms of the GNU General Public
 * License.  See COPYING or http://www.gnu.org/ for more details.
 *
 * Routines for manipulation of keysyms, keycodes, keyboard remapping
 * and such.
 */
#define KMANIP_C

#include "master.h"

#ifndef PRINT_STATUS_MESSAGES
#  define PRINT_STATUS_MESSAGES
#endif /* PRINT_STATUS_MESSAGES */

/*************** FUNCTIONS ****************************/
void remap_at_exit_dialog(GtkWidget *emitter, gpointer data)
{

} /* End remap_at_exit_dialog() */

/* Wrapper for gtkeyboard_remap_keyboard() that destroys popup windows
 * requesting the action.
 */
void remap_keyboard_wrapper(GtkWidget *emitter, gpointer data)
{
     if(GTK_IS_DIALOG(emitter->parent->parent))
	  gtk_widget_destroy(emitter->parent->parent);
     else if(GTK_IS_DIALOG(emitter->parent->parent->parent))
	  gtk_widget_destroy(emitter->parent->parent->parent);
     
     return;
} /* End remap_keyboard_wrapper() */

void remap_default_at_exit(GtkWidget *emitter, gpointer data)
{

} /* End remap_default_at_exit */

/* Remaps the keyboard completely to whatever is in options.keyboard */
void gtkeyboard_remap_keyboard(GtkWidget *emitter, gpointer data)
{
     GtkWidget *e = emitter;
     KEY *key;
     int x, y;
     char error_buffer[1024 * 5];
     KeySym symarr[3];
     int symbols_to_remap;

     if(emitter)
     {
	  while(e && GTK_IS_WIDGET(e) && !GTK_IS_WINDOW(e))
	       e = e->parent;
     }

     if(!options.keyboard)
     {
	  sprintf(error_buffer,
		  "Error:  Can't remap keyboard:\n");
	  strcat(error_buffer,
		 "No keyboard definition loaded.\n");
	  annoying_popup(error_buffer);
	  return;
     } /* End if */

     /* We have to do a dry run through all of the keys to check and see if
      * they have valid KeyCodes specified.  We don't want to find out half
      * way through the mapping that we can't finish the mapping.
      * Report the error and return if there is even one key without a valid
      * KeyCode.
      */
     for(x=0; x<MAXIMUM_ROWS; x++)
     {
	  for(y=0; y<options.keyboard->row_values[x]; y++)
	  {
	       key = gtkeyboard_keyboard_get_key(options.keyboard, x, y);
	       
	       if(!key->code)
	       {
		    sprintf(error_buffer,"Error:  Cannot remap keyboard:\n%s",
			    options.keyboard->name); 
		    strcat(error_buffer,"Some keycodes are missing.\n");

		    if(e && GTK_IS_WIDGET(e))
			 gtk_widget_destroy(e);

		    annoying_popup(error_buffer);
		    return;
	       } /* End if */

	       key = gtkeyboard_destroy_key(key);
	  } /* End for */
     } /* End for */

     /* Now do the mapping for real.  Just map each keycode as the keyboard
      * specifies to the set: lower_case, upper_case, alt_gr
      */
     for(x=0; x<MAXIMUM_ROWS; x++)
     {
	  for(y=0; y<options.keyboard->row_values[x]; y++)
	  {
	       /* Grab our current key */
	       key = gtkeyboard_keyboard_get_key(options.keyboard, x, y);

	       symarr[0] = key->lower_case;
	       symarr[1] = key->upper_case;
	       symarr[2] = key->alt_gr;

	       if(symarr[2] == NoSymbol)
		    symbols_to_remap = 2;
	       else symbols_to_remap = 3;

	       /* Remap one KeyCode using symarr, map it to symbols_to_remap
		* different symbols.
		*/
	       XChangeKeyboardMapping(GDK_DISPLAY(), key->code, 
				      symbols_to_remap, &symarr[0], 1);
	  } /* End for */
     } /* End for */

     /* Update before exit */
     XFlush(GDK_DISPLAY());

     if(e && GTK_IS_WIDGET(e))
	  gtk_widget_destroy(e);

     if(options.keyboard->name)
	  sprintf(error_buffer,"Remapping to\n%s\nSuccessful!\n",
		  options.keyboard->name);
     else
	  sprintf(error_buffer,"Remapping successful!\n");

     annoying_popup(error_buffer);
} /* End gtkeyboard_remap_keyboard() */





